/*╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌
 * qtModels/tableModels/BoilTableModel.h is part of Brewtarget, and is copyright the following authors 2024-2025:
 *   • Matt Young <mfsy@yahoo.com>
 *
 * Brewtarget is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Brewtarget is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 ╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌*/
#ifndef TABLEMODELS_BOILTABLEMODEL_H
#define TABLEMODELS_BOILTABLEMODEL_H
#pragma once

#include <QItemDelegate>

#include "model/Boil.h"
#include "qtModels/tableModels/BtTableModel.h"
#include "qtModels/tableModels/ItemDelegate.h"
#include "qtModels/tableModels/TableModelBase.h"

// Define the columns on this table
TABLE_MODEL_TRAITS(Boil, Name            ,
                         PreBoilSize     ,
                         NumRecipesUsedIn,)

/*!
 * \class BoilTableModel
 *
 * \brief Model class for a list of boils.
 */
class BoilTableModel : public BtTableModel, public TableModelBase<BoilTableModel, Boil> {
   Q_OBJECT
   TABLE_MODEL_COMMON_DECL(Boil)
};

/**
 * \class BoilItemDelegate
 *
 * \brief An item delegate for boil tables.
 * \sa BoilTableModel
 */
class BoilItemDelegate : public QItemDelegate, public ItemDelegate<BoilItemDelegate, BoilTableModel> {
   Q_OBJECT
   ITEM_DELEGATE_COMMON_DECL(Boil)
};

#endif
